
/* Portions Copyright (c) 1995,1996 by Trusted Information Systems, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND TRUSTED INFORMATION SYSTEMS DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TRUSTED INFORMATION 
 * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 *
 * Trusted Information Systems, Inc. has received approval from the
 * United States Government for export and reexport of TIS/DNSSEC
 * software from the United States of America under the provisions of
 * the Export Administration Regulations (EAR) General Software Note
 * (GSN) license exception for mass market software.  Under the
 * provisions of this license, this software may be exported or
 * reexported to all destinations except for the embargoed countries of
 * Cuba, Iran, Iraq, Libya, North Korea, Sudan and Syria.  Any export
 * or reexport of TIS/DNSSEC software to the embargoed countries
 * requires additional, specific licensing approval from the United
 * States Government. 
 */




static char *rcsid = "$Header: /u/h/ogud/sec_bind/dnssec_lib/RCS/dnssec_api.c,v 1.18 1996/09/17 20:39:41 ogud Exp $";
/*
 * This file contains the interface between the DNSSEC API to the crypto 
 * API. 
 * This is the only file that needs to be changed if the crypto system is 
 * changed.
 * exported functions are 
 * int  dnssec_check()     Function to check if certain alg is suppored
 * int  dnssec_sign():     Incremental signing routine takes flags to 
 *                         select which steps to perform 
 * int  dnssec_verify()    Incremental verify routine
 * KEY *dnssec_pubkey()    Function to convert KEY RR into a KEY structure
 * KEY *dnssec_getkey()    Function to retrive private KEY from storage
 *                         private keys are stored internaly and are only 
 *                         decoded once.
 * int  dnssec_genkey()    Function to generate new KEY for dns
 * void dnssec_putkey()    Function to write out a private key ??
 * void dnssec_freekey()   Releases all memory pointed to by key structure
 * 
 */


#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h> 
#include <netinet/in.h>  
#include "../include/arpa/nameser.h"
#include "../conf/options.h"
#ifdef DNS_SECURITY
#include "../conf/portability.h"
#include "pathnames.h"
#include "ns_defs.h"
#include "db_defs.h" 
#include "dnssec.h" 
#include "dns_crypto.h"

#ifdef TESTING 
#define DUMP(data, len, msg){   char x[1000];\
				int n=res_encode64( data, x, len);\
				printf("%s: %d %s\n",msg,n, x);}
#else 
#define DUMP(a,b,c)
#endif  /* TESTING */
#define DES_COMPAT 0

static int    dnssec_decrypt    __P((u_char *, u_char *, int));
static int    dnssec_encrypt    __P((u_char *, u_char *, int));
static void   dnssec_putkey     __P((KEY *));

static int    ReadPrivateKey    __P((char *, KEY *));
static int    WritePublicKey    __P((KEY *, int));
/* 
 * dnssec_check()  Function to check if crypto system for specified 
 *   algorithm is present. 
 * Return 1 if it is 
 *        0 if not present or not supported 
 */
int 
dnssec_check( alg)
int alg; 
{
  switch( alg) {
    case KEY_RSA:
#ifdef RSAREF
       return 1; 
#endif /* RSAREF */
       break;
    default:
       break;
 /* add other supported alg here */
  }
  return 0; 
}

/*
 * Incremental signing function 
 * If len and data are non NULL the contents of data is added to the hash
 * even if the flags are INIT or FINAL
 * Parameters: 
 *     mode  what operations to perform this time. 
 *     data  the data to pass through the hash function
 *     len   the length of the data 
 *     in_key the key to use 
 *     signature the location to write the signature to 
 * Return codes
 *    0 success INIT or Update operation 
 *    positive value success FINAL (sign)  operation
 *    negative value failure 
 */ 

int 
dnssec_sign(mode, data, len,  in_key, signature)
int mode, len;
u_char *data;
KEY *in_key;
u_char *signature;
{
  int sign_len=0;
  int axfr =  mode & SIG_MODE_AXFR ? 1 : 0;
  if(axfr && len > 0 && data ) DUMP(data, len, "dnssec_sign()");
  switch (in_key->alg) {
    case KEY_RSA: 
#ifdef RSAREF
    {
      RSA_Key *key;
      key = (RSA_Key *) in_key->KEY_struct;
      sign_len = rsa_sign( mode, &key->context[axfr], data, len, 
			  key->Private_Key,   signature);
      return sign_len;
    }
#endif /*RSAREF */
    default: 
         break;
  }
  return UNKNOWN_KEYALG;
}
/*
 * Incremental verification routine
 * This routine will add data to the hash on every call if any data is passed
 * in even when the mode is INIT or FINAL 
 * Parameters: 
 *     mode  what operations to perform this time. 
 *     data  the data to pass through the hash function
 *     len   the length of the data 
 *     in_key the key to use 
 *     signature the location to write the signature to 
 *     sig_len  lenght of the signature in bytes.
 * Return codes
 *    0 success 
 *    any other value failure 
 */
int 
dnssec_verify(mode, data, len,  in_key, signature, sig_len)
int mode,len, sig_len;
u_char *data;
KEY *in_key;
u_char *signature;
{
  int out=1;
  int axfr =  mode & SIG_MODE_AXFR ? 1: 0;
  if( axfr && data && len > 0) DUMP(data, len, "dnssec_verify()");
  switch (in_key->alg) {
  case KEY_RSA: 
#ifdef RSAREF
    {
      RSA_Key *key = (RSA_Key *) in_key->KEY_struct;
      return rsa_verify( mode, &key->context[axfr], data, len, 
			  key->Public_Key, signature, sig_len);
    }
#endif /* RSAREF */
    default:
      break;
  }
  return UNKNOWN_KEYALG;
}

/*
 * Key Functions
 * Function to get private key 
 * Parameters 
 *      keyname  the name of the KEY
 * Return
 *      NULL     if there is no key found in the current directory 
 *               or this key has not been loaded before 
 *     !NULL     KEY structure filled out 
 */
KEY * 
dnssec_getkey( keyname ) 
char * keyname;
{
  KEY *dg_key=NULL; 
  int flags, t;
  FILE *fp;
  char str[MAXDNAME];
  static struct key_chain {
    KEY *key;
    struct key_chain *next; 
  } *chain_head=NULL, *next_key; 
  /* look for key among keys that have already been read in,  I need to
     purge the chain once in a while,  each key needs an expiration time
  */ 
  if(keyname == NULL) return NULL; 
  t =strlen(keyname);  /* make sure no keyname ends with a . */
  if( keyname[t-1] == '.') keyname[--t] = '\0';
  for( next_key = chain_head; next_key && next_key->key; 
      next_key=next_key->next)
    if(strcasecmp( keyname, next_key->key->signer)==0 && 
       t == strlen(next_key->key->signer)) break;

  if( next_key)  /* found the key */ 
    return next_key->key;
  /* before I read in the key check if it is allowed to sign */ 
  (void) sprintf(str,"%s.public",keyname);
  if((fp = fopen(str,"r")) == NULL){
    printf("dnssec_getkey(): Public Key not found %s\n", str);
    return NULL;
  }
  if(fscanf(fp,"%d",&flags) != 1) { 
    printf("dnssec_getkey(): Can not read flags field from %s\n",str);
    return NULL;
  }
  fclose(fp);
  if( ((flags &(SIGN_F_SIGNATORY | SIGN_F_ZONE)) == 0) || /* allowed to sign*/
      ((flags & SIGN_F_NOAUTHENT) == SIGN_F_NOAUTHENT)) { 
    printf("dnssec_getkey():Key %s not allowed to sign or authenticate %x\n",
	   keyname, flags);
    return NULL;
  }
  dg_key = (KEY *) malloc(sizeof(KEY));
  if( ReadPrivateKey( keyname, dg_key)) {
    dg_key->axfr = do_axfr;
    dg_key->flags = flags;
    dg_key->ttl   = 1;  /* I need to set this to something better later XXXX*/

    /*
     * add this key to the chain of existing keys  we add to front as we
     * expect that each key will only be used for short time. 
     * and the chain will be short 
     */
    next_key = (struct key_chain *) malloc( sizeof(struct key_chain)); 
    next_key->key = dg_key; 
    next_key->next = chain_head; 
    chain_head = next_key;
  }
  else {
    SAFE_FREE(dg_key);
    dg_key = NULL;
  }
  return dg_key;
}


/*
 * Generate a public/private key pair and store them
 * Parameters 
 *     name   the name of the new key 
 *     bits   the size of the new key in bits 
 *     exp    what exponent to use  0 ==> 3  any other value Fermant4
 *     flags  what DNS KEY flags this key has by default 
 *     protocol the default value of the DNS KEY protocol field
 *     alg    what KEY algorithm to use 
 * Return codes
 *     0 Failure 
 *     1 Success
 */
int
dnssec_genkey(name, bits, exp, flags, protocol, alg)
int bits, exp, flags, protocol, alg;
char *name;
{
  KEY new_key;
  new_key.signer   = name;
  new_key.key_size = bits;
  new_key.alg      = alg;
  new_key.flags    = flags;
  new_key.KEY_struct= NULL;
  if( (flags & SIGN_F_NOKEY) == SIGN_F_NOKEY){ 
    WritePublicKey(&new_key, protocol);
  }
  else {
    switch( alg) {
    case KEY_RSA:
#ifdef RSAREF
      { 
	if(  Generate_RSAKey(&new_key,exp) == 0){
	  printf("dnssec_genkey(): Key generation failure %s %d %d %d\n",
		 new_key.signer, new_key.alg, new_key.key_size, exp);
	  return 0;
	}
	break;
      }
#endif /* RSAREF */
    default: 
      printf("dnssec_genkey():Unsupported algorithm %d\n",alg);
      return 0;
    }
  }
  dnssec_putkey( &new_key);
  WritePublicKey( &new_key, protocol);
  return 1;
}

/*
 * Function to write out private KEY
 * This function is should only be called by dnssec_genkey() but is included
 * here to allow future extensions such as changing password of keys
 * Parameters: 
 *    name     the name of the File to store the KEY in
 *    private  The Private key to store 
 *    size     Size of the private key
 *    alg      What algorithm this is
 */

static void 
dnssec_putkey( key)
KEY  *key;
{
  u_char encoded_block[RAW_KEY_SIZE],encrypted_block[RAW_KEY_SIZE];
  char file[RAW_KEY_SIZE];
  int i,len, out_len; 
  /* First I need to encode the key into the portable key format */
  switch ( key->alg){
   case KEY_RSA:
#ifdef RSAREF
    { 
      if( (len = Encode_RSAKey( ((RSA_Key *) key->KEY_struct)->Private_Key,
			      encoded_block, sizeof(encoded_block))) <= 0){
	printf( "dnssec_putkey(): Could not encode private RSA key %d\n", len);
	return ;
      }
      break;
    }
#endif /* RSAREF */
  default: 
    printf("dnssec_putkey(): Unknown algorithm %d\n", key->alg);
    return;
  }
  /* Now I can create the file I want to use */
  sprintf(file,"%s.priv",key->signer);
  if( rename_old( file,
	     "dnssec_putkey():Old Private key file exists Renaming to %s\n"))
      printf( "dnssec_putkey(): Rename Falied %d\n", errno);

  if( (out_len =dnssec_encrypt( encoded_block, encrypted_block, len))>0){
    if( (i=open(file, O_WRONLY| O_CREAT | O_TRUNC, 0700))) {
      int nn;
      if((nn=write(i, encrypted_block, out_len)) != out_len) { 
	printf("dnssec_putkey(): Write failure on %s %d != %d errno=%d\n",
	       file, out_len, nn, errno);
	exit(-3);
      }
      close(i);
    }
    else 
      printf("dnssec_putkey(): Can not create file %s\n",file);
  }
  else 
      printf("dnssec_putkey(): Encryption failed %d %d\n", out_len,len);
  memset(encoded_block, 0, len);
  memset(encrypted_block, 0, len);
}

/*
 * Following are functions that actualy deal with low level stuff of keys
 * and know about key formats
 *
 * ReadPrivateKey( name, KEY structure) 
 *      Function reads in private key with from a file named <name>.priv 
 *      Fills out the KEY structure 
 *  return 1 if everthing works 
 *         0 if there is any problem
 *
 */
static int
ReadPrivateKey (name, pk_key)
     char *name;
     KEY *pk_key;
{
  int fd, cnt, alg, len;
  int32_t foot;
  char filename[200];
  u_char in_buff[RAW_KEY_SIZE], out_buff[RAW_KEY_SIZE], *p;
  float x;

  /* first check if we can find the key file */ 
  if( strlen(name)<5 || memcmp(&name[strlen(name)-5], ".priv",5))
    sprintf( filename, "%s.priv",name);
  else 
    strcpy( filename, name);
  if ((fd = open(filename, O_RDONLY, 0)) == -1) { 
    printf( "ReadPrivateKey: Could not open file %s in directory %s\n",
	   filename,(char *) getcwd(NULL, MAXPATHLEN-1));
    return 0;
  }
  /* now read the header info from the file */
  if( (cnt=read(fd, in_buff, sizeof(in_buff))) < 5) {
    close(fd);
    printf("ReadPrivateKey: error reading file %s(empty file)\n", filename);
    return 0;
  }
  close(fd);
  printf("Decoding Key %s\n", name);
  if(memcmp(in_buff, "Private-key-format: v",20)!=0) {  /* File encrypted */
    /*simulate des program */
    if( (len=dnssec_decrypt(in_buff, out_buff, cnt)) <= 0){
      printf("ReadPrivateKey():Failed decrypting key %s\n",name);
      goto fail;
    }
  }
  else 
    memcpy( out_buff, in_buff, len=cnt);
  p=out_buff;
  if (!verify_str((char **) &p, "Private-key-format: v")) {
    printf("ReadPrivateKey(): Not a Key file/Decrypt failed %s\n", name);
    goto fail;
  }
  sscanf(p,"%f", &x);  /* read in file format */
  if( x < 1.0) {
    printf("ReadPrivateKey(): Unknown keyfile %f version for %s\n",
	   x, name);
    goto fail;
  }
  else if( x > (float)atof(KEY_FILE_FORMAT)) {
    printf(
      "ReadPrivateKey(): Keyfile %s version higher than mine %f MAY FAIL\n",
	   name,x);
  }
  while( *p++!='\n');  /* skip to end of line */
  if(!verify_str((char **) &p, "Algorithm: "))
    goto fail;
  if( sscanf(p,"%d", &alg) != 1) 
    goto fail;
  while(*p++ != '\n'); /* skip to end of line */
  pk_key->signer = (char *) strdup( name);
  
  switch ( alg) { 
  case KEY_RSA:
#ifdef RSAREF 
    {
      RSA_Key *r_key = (RSA_Key *) malloc( sizeof(RSA_Key ));
      memset(r_key, 0, sizeof(RSA_Key));
      foot = Decode_RSAKey(r_key, p, &out_buff[len]-p);
      if( foot < 0 || r_key->Private_Key == NULL) { 
	printf("ReadPrivateKey():Decode_RSAKey failed \n");
	memset(r_key,0, sizeof(RSA_Key));
	SAFE_FREE(r_key);
	goto fail;
      }
      pk_key->key_size = r_key->Private_Key->bits;
      r_key->signer =pk_key->signer; 
      pk_key->foot = (u_int16_t) foot;
      pk_key->alg  = alg;
      pk_key->KEY_struct = r_key;
      break;
    }
#endif /* RSAREF */
  default: 
    printf("ReadPrivateKey():Unknown Key type %x\n", alg);
    goto fail;
  }
  memset( in_buff, 0, cnt);
  memset( out_buff,0, cnt); 
  return 1;
fail:
  memset( in_buff,  0, cnt);
  memset( out_buff, 0, cnt);
  return 0;
}

/* 
 * dnssec_pubkey() Function to convert contents of an KEY RR into a 
 *                 key structure
 * This routine gets passed a name of the key, the algorithm and the beginning
 * of the key data in the KEY RR (byte 4). 
 * It returns a key structure  or NULL if unsuccessfull
 */
KEY * 
dnssec_pubkey(alg, len, key, name)
int alg, len;
u_char *key, *name;
{
  KEY *key_st = (KEY *) malloc( sizeof(KEY));
  key_st->alg = alg;
  key_st->axfr = 0;
  
  switch( alg) { 
    case KEY_RSA:
#ifdef RSAREF  
    {
      dns2rsa_key(len, key, name,key_st);
      break;
    }
#endif /* RSAREF */
  default: 
    printf("dnssec_pubkey(): Unknown alg %d\n",alg);
    SAFE_FREE(key_st);
    key_st = NULL;
  }
  return key_st;
}

/* 
 * dnssec_freekey( key) 
 * function to release all data structures pointed to by a key record
 * and the key record passed in 
 */ 
void
dnssec_freekey( f_key)
KEY *f_key;
{
  if(f_key== NULL) 
    return;
  switch(f_key->alg) {
    case KEY_RSA: 
#ifdef RSAREF
      {
	Free_RSAKey((RSA_Key *) f_key->KEY_struct);
	f_key->KEY_struct = NULL;
	break; 
      }
#endif /* RSAREF */
    default: 
      printf("dnsfree_key(): Unknown key alg %d\n",f_key->alg);
  }
  if( f_key->KEY_struct) SAFE_FREE((char *)f_key->KEY_struct);
  if( f_key->signer)     SAFE_FREE(f_key->signer);
  SAFE_FREE(f_key);
}

/* 
 * dnssec_decrypt function to decrypt the content of the private key file 
 * inputs: ibuf == buffer with encrypted file contents
 *         obuf == buffer to put decrypted file into 
 *         len  == number of bytes in ibuf 
 */ 
static int 
dnssec_decrypt( ibuf, obuf, len)
     u_char *ibuf, *obuf;
     int len;
{
  int tries=3, ret =0, j;
  u_char pass[40];
  char *str="Private-key-format: v";
  while (tries-- && 
	 (ret=get_passwd(pass, "Key decryption Password:", sizeof(pass)))>=0){
    if(ret) des_function( pass, ibuf, obuf, len, 0); /* mode is decrypt */
    else memcpy(obuf,ibuf, len);
    if(!(ret=strncmp(obuf, str, strlen(str)))) {
      memset(pass,0,sizeof(pass));
      j=(int) obuf[len-1];
      return len-j;
    }
  }
  printf("Dnssec_decrypt():Failed decrypting the buffer\n");
  memset(pass,0,sizeof(pass));
  memset(obuf, 0, len);
  return 0;
}
/* 
 * dnssec_encrypt function to encrypt the content of the private key file 
 * inputs: ibuf == buffer with encrypted file contents
 *         obuf == buffer to put encrypted file into 
 *         len  == number of bytes in ibuf 
 */ 
static int 
dnssec_encrypt( ibuf, obuf, len)
     u_char *ibuf, *obuf;
     int len;
{
  u_char pass[2] [40];
  int tries=3, p1,p2;
  memset(pass, 0, sizeof(pass));
  printf("Enter Encryption password for the key, <Return> for no password\n");
  for( tries=3; tries==3 ||(tries >= 0 && strcmp( pass[1], pass[0])); tries--){
    p1 = get_passwd(pass[0], "Key Encryption Password:", sizeof(pass[1]));
    p2 = get_passwd(pass[1], "Retype Key Encryption Password:", 
		    sizeof(pass[1]));
  }
  if( p1 == p2 && !strncmp(pass[1], pass[0], p1)){
    if( p1 > 0 )
      len = des_function( pass[0], ibuf, obuf, len, 1);
    else  /* no password */
      memcpy( obuf, ibuf, len);
    memset(pass, 0, sizeof(pass));
  }
  else { 
    printf("dnssec_encrypt(): Non acceptable password\n");
    return 0;
  }
 return len;
}

/* 
 * WritePublicKey() function to write to file public key in DNS format
 * Inputs: 
 *    Key : a generic Key structure
 *    proto: The value of the proto field for the public KEY
 * Output:
 *    0 if failed 
 *    1 if successful
 */ 
static int
WritePublicKey( key, proto) 
KEY *key;
int proto;
{
  FILE *fp;
  char filename[RAW_KEY_SIZE];
  char out_key[RAW_KEY_SIZE];
  int len;

  if(key ==NULL || key->KEY_struct == NULL) { 
    printf("WritePublicKey(): No Key specified \n");
    return 0; 
  }

  memset(out_key, 0, sizeof(out_key));
  switch( key->alg) {
    case KEY_RSA: 
#ifdef RSAREF 
      len = rsa2dns_key( ((RSA_Key *)key->KEY_struct)->Public_Key, out_key);
      break;
#endif
    default:
      len = 0;
      printf("Trying to convert an unkown algorithm %d\n", key->alg);
      return;
  }
      /* create public key file */
  sprintf(filename, "%s.public",key->signer);
  if( rename_old(filename, "WritePublicKey(): renamed old key %s\n"))
    printf("WritePublicKey():rename_old Failed %s",filename);

  if ((fp = fopen(filename, "w+")) == NULL) {
    printf("Write_PublicKeys:open of file:%s failed (errno=%d)\n", 
	   filename, errno);
    exit(-12);
  }
           /*write out key */
  fprintf(fp,"%d %d %d %s\n",key->flags, proto, key->alg,out_key);
  fclose(fp);
  return 1;
}
#endif /* DNS_SECURITY */
